import numpy
from collections import OrderedDict
import pickle

def bin_ndarray(ndarray, new_shape, weights=None, operation=numpy.mean):
    """
    Bins an ndarray in all axes based on the target shape, by summing 
    or averaging.
    
    Parameters
    ----------
    ndarray : array_like
        the input array to re-bin
    new_shape : tuple
        the tuple holding the desired new shape
    weights : array_like, optional
        weights to multiply the input array by, before running the re-binning
        operation, 
    
    Notes
    -----
    *   Dimensions in `new_shape` must be integral factor smaller than the 
        old shape
    *   Number of output dimensions must match number of input dimensions.
    *   See https://gist.github.com/derricw/95eab740e1b08b78c03f
    
    Example
    -------
    >>> m = numpy.arange(0,100,1).reshape((10,10))
    >>> n = bin_ndarray(m, new_shape=(5,5), operation=numpy.sum)
    >>> print(n)
    [[ 22  30  38  46  54]
     [102 110 118 126 134]
     [182 190 198 206 214]
     [262 270 278 286 294]
     [342 350 358 366 374]]
    """
    if ndarray.shape == new_shape:
        raise ValueError("why are we re-binning if the new shape equals the old shape?")
    if ndarray.ndim != len(new_shape):
        raise ValueError("Shape mismatch: {} -> {}".format(ndarray.shape, new_shape))
    if numpy.sometrue(numpy.mod(ndarray.shape, new_shape)):
        args = (str(new_shape), str(ndarray.shape))
        msg = "desired shape of %s must be integer factor smaller than the old shape %s" %args
        raise ValueError(msg)
        
    compression_pairs = [(d, c//d) for d, c in zip(new_shape, ndarray.shape)]
    flattened = [l for p in compression_pairs for l in p]
    ndarray = ndarray.reshape(flattened)
    if weights is not None: weights = weights.reshape(flattened)
    
    for i in range(len(new_shape)):
        if weights is not None:
            ndarray = operation(ndarray*weights, axis=-1*(i+1))
            weights = numpy.sum(weights, axis=-1*(i+1))
            ndarray /= weights
        else:
            ndarray = operation(ndarray, axis=-1*(i+1))
    return ndarray

def __unpickle__(cls, d): 
    """
    Internal 'unpickling' function to create a new `cls` instance 
    from the dictionary `d`, which stores all the necessary information. 
    
    Notes
    -----
    *   Designed to be used with `DataSet.__reduce__`
    """
    data, mask = d.pop('data'), d.pop('mask')
    return cls.__construct_direct__(data, mask, **d)    
    
class DataSet(object):
    """
    Lightweight class to hold variables at fixed coordinates, i.e.,
    a grid of (r, mu) or (k, mu) bins for a correlation function or
    power spectrum measurement
    
    It is modeled after the syntax of :class:`xarray.Dataset`, and is designed 
    to hold correlation function or power spectrum results (in 1D or 2D)
    
    Notes
    -----
    *   the suffix *cen* will be appended to the names of the 
        dimensions passed to the constructor, since the :attr:`coords` array holds
        the **bin centers**, as constructed from the bin edges
        
    Examples
    --------
    The following example shows how to read a power spectrum or correlation 
    function measurement as written by a nbodykit `Algorithm`. It uses
    :func:`~nbodykit.files.Read1DPlainText`
    
    >>> from nbodykit import files
    >>> corr = Corr2dDataSet.from_nbkit(*files.Read2DPlainText(filename))
    >>> pk = Power1dDataSet.from_nbkit(*files.Read1DPlainText(filename))
        
    Data variables and coordinate arrays can be accessed in a dict-like
    fashion:
        
    >>> power = pkmu['power'] # returns power data variable
    >>> k_cen = pkmu['k_cen'] # returns k_cen coordinate array
        
    Array-like indexing of a :class:`DataSet` returns a new :class:`DataSet`
    holding the sliced data:
        
    >>> pkmu
    <DataSet: dims: (k_cen: 200, mu_cen: 5), variables: ('mu', 'k', 'power')>
    >>> pkmu[:,0] # select first mu column
    <DataSet: dims: (k_cen: 200), variables: ('mu', 'k', 'power')>
    
    Additional data variables can be added to the :class:`DataSet` via:
    
    >>> modes = numpy.ones((200, 5))
    >>> pkmu['modes'] = modes
    
    Coordinate-based indexing is possible through :func:`sel`:
    
    >>> pkmu
    <DataSet: dims: (k_cen: 200, mu_cen: 5), variables: ('mu', 'k', 'power')>
    >>> pkmu.sel(k_cen=slice(0.1, 0.4), mu_cen=0.5)
    <DataSet: dims: (k_cen: 30), variables: ('mu', 'k', 'power')>
    
    :func:`squeeze` will explicitly squeeze the specified dimension 
    (of length one) such that the resulting instance has one less dimension:
        
    >>> pkmu
    <DataSet: dims: (k_cen: 200, mu_cen: 1), variables: ('mu', 'k', 'power')>
    >>> pkmu.squeeze(dim='mu_cen') # can also just call pkmu.squeeze()
    <DataSet: dims: (k_cen: 200), variables: ('mu', 'k', 'power')>
    
    :func:`average` returns a new :class:`DataSet` holding the 
    data averaged over one dimension
        
    :func:`reindex` will re-bin the coordinate arrays along the specified 
    dimension
    """
    _fields_to_sum = []
    
    def __init__(self, dims, edges, variables, **kwargs):
        """
        Parameters
        ----------
        dims : list, (Ndim,)
            A list of strings specifying names for the coordinate dimensions.
            The dimension names stored in :attr:`dims` have the suffix 'cen'
            added, to indicate that the coordinate grid is defined at the bin 
            centers
            
        edges : list, (Ndim,)
            A list specifying the bin edges for each dimension
            
        variables : dict
            a dictionary holding the data variables, where the keys
            are interpreted as the variable names. The variable names are 
            stored in :attr:`variables`
            
        **kwargs :
            Any additional keywords are saved as metadata in the :attr:`attrs`
            attribute, which is an :class:`~collections.OrderedDict`
        """
        if len(dims) != len(edges):
            raise ValueError("size mismatch between specified `dims` and `edges`")
        
        shape = tuple(len(e)-1 for e in edges)
        for name in variables:
            if numpy.shape(variables[name]) != shape:
                args = (shape, numpy.shape(variables[name]))
                raise ValueError("`edges` imply data shape of %s, but data has shape %s" %args)
        
        self.dims = [dim+'_cen' for dim in dims]
        self.edges = dict(zip(self.dims, edges))
        
        # coordinates are the bin centers
        self.coords = {}
        for i, dim in enumerate(self.dims):
            self.coords[dim] = 0.5 * (edges[i][1:] + edges[i][:-1])
            
        # store variables as a structured array
        dtypes = [(name, variables[name].dtype) for name in variables]            
        self.data = numpy.empty(self.shape, dtype=numpy.dtype(dtypes))
        for name in variables:
            self.data[name] = variables[name]
            
        # define a mask such that a coordinate grid element will be masked
        # if any of the variables at that coordinate are (NaN, inf)
        self.mask = numpy.zeros(self.shape, dtype=bool)
        for name in variables:
            self.mask = numpy.logical_or(self.mask, ~numpy.isfinite(self.data[name]))
             
        # save and track metadata
        self.attrs = OrderedDict()
        for k in kwargs: self.attrs[k] = kwargs[k]
        
    @property
    def shape(self):
        """
        The shape of the coordinate grid
        """
        return tuple(len(self.coords[d]) for d in self.dims)
        
    @property
    def variables(self):
        """
        Alias to return the names of the variables stored in `data`
        """
        return list(self.data.dtype.names)
                
    @classmethod
    def __construct_direct__(cls, data, mask, **kwargs):
        """
        Shortcut around __init__ for internal use to construct and
        return a new class instance. The returned object should be 
        identical to that returned by __init__.
        
        Notes
        -----
        *   Useful for returning new instances with sliced data/mask 
        *   The keyword arguments required to create a full, unbroken
            instance are `dims`, `coords`, `edges`, and `attrs`
                    
        Parameters
        ----------
        data : 
        """
        obj = object.__new__(cls)
        for k in kwargs: setattr(obj, k, kwargs[k])
            
        for k, d in zip(['data', 'mask'], [data, mask]):
            setattr(obj, k, d)
            if obj.shape != d.shape:
                try:
                    setattr(obj, k, d.reshape(obj.shape))
                except:
                    raise ValueError("shape mismatch between data and coordinates")
        return obj
    
    def __copy_attrs__(self):
        """
        Return a copy of all necessary attributes associated with
        the `DataSet`. This dictionary + `data` and `mask` are all
        that's required to reconstruct a new class
        """
        kw = {}
        kw['dims'] = list(self.dims)
        kw['edges'] = self.edges.copy()
        kw['coords'] = self.coords.copy()
        kw['attrs'] = self.attrs.copy()
        kw['_fields_to_sum'] = list(self._fields_to_sum)
        return kw

    def __finalize__(self, data, mask, indices):
        """
        Finalize and return a new instance from a slice of the 
        current object (returns a copy)
        """
        edges, coords = self.__slice_edges__(indices)
        kw = {'dims':list(self.dims), 'edges':edges, 'coords':coords, 
                'attrs':self.attrs.copy(), '_fields_to_sum':self._fields_to_sum}
        return self.__class__.__construct_direct__(data, mask, **kw)

    def __slice_edges__(self, indices):
        """
        Internal function to slice the `edges` attribute with the 
        specified indices, which specify the included coordinate bins
        """
        edges = {}
        coords = {}
        for i, dim in enumerate(self.dims):
            idx = indices[i] + [indices[i][-1]+1]
            edges[dim] = self.edges[dim][idx]
            coords[dim] = 0.5 * (edges[dim][1:] + edges[dim][:-1])
        
        return edges, coords
        
    def __str__(self):
        name = self.__class__.__name__
        dims = "(" + ", ".join(['%s: %d' %(k, self.shape[i]) for i, k in enumerate(self.dims)]) + ")"
        
        if len(self.variables) < 5:
            return "<%s: dims: %s, variables: %s>" %(name, dims, str(tuple(self.variables)))
        else:
            return "<%s: dims: %s, variables: %d total>" %(name, dims, len(self.variables))
        
    def __repr__(self):
        return self.__str__()
        
    def __contains__(self, key):
        return key in self.variables
    
    def __setitem__(self, key, data):
        """
        Add a new variable with the name `key` to the class using `data`
        """
        if numpy.shape(data) != self.data.shape:
            raise ValueError("data to be added must have shape %s" %str(self.data.shape))
            
        # add the new (key, type) to the dtype, or if the key is present, overwrite
        dtype = list(self.data.dtype.descr) # make copy
        names = list(self.data.dtype.names) # make copy
        if key in names:
            i = names.index(key)
            dtype.pop(i); names.pop(i)
        dtype += [(key, data.dtype.type)]
        
        # add old variables
        new = numpy.zeros(self.data.shape, dtype=dtype)
        for col in names:
            new[col] = self.data[col]
   
        # the new data to add
        new[key] = data
        mask = numpy.logical_or(self.mask, ~numpy.isfinite(new[key]))
        
        # save
        self.data = new
        self.mask = mask
        
    def __getitem__(self, key):
        """
        Index- or string- based indexing
        
        Notes
        -----
        *   If a single string is passed, the key is intrepreted
            as a `variable` or `coordinate`, and the corresponding
            array is returned
        *   If a list of strings is passed, then a new `DataSet`
            holding only the `variable` names in `key` is returned
        *   Integer-based indexing or slices similar to numpy 
            indexing will slice `data`, returning a new 
            `DataSet` holding the newly sliced data and coordinate grid
        *   Scalar indexes (i.e., integers) used to index a certain
            dimension will "squeeze" that dimension, removing it
            from the coordinate grid
        """
        # if single string passed, return a coordinate or variable
        if isinstance(key, str):
            if key in self.variables:
                return self.data[key]
            elif key in self.coords.keys():
                return self.coords[key]
            else:
                raise KeyError("`%s` is not a valid variable or coordinate name" %key)
            
        # indices to slice the data with
        indices = [range(0, self.shape[i]) for i in range(len(self.dims))]
            
        # check for list/tuple of variable names
        # if so, return a DataSet with slice of columns
        if isinstance(key, (list, tuple)) and all(isinstance(x, str) for x in key):
            if all(k in self.variables for k in key):
                return self.__finalize__(self.data[key], self.mask.copy(), indices)
            else:
                invalid = ', '.join("'%s'" %k for k in key if k not in self.variables)
                raise KeyError("cannot slice variables -- invalid names: (%s)" %invalid)
            
        key_ = key
        
        # if key is a single integer or single slice, make it a list
        make_iterable = isinstance(key, slice) or isinstance(key, int)
        if make_iterable or isinstance(key, list) and all(isinstance(x, int) for x in key):
            key_ = [key]

        squeezed_dims = []
        for i, subkey in enumerate(key_):
            if i >= len(self.dims):
                raise IndexError("too many indices for DataSet; note that ndim = %d" %len(self.dims))
                
            if isinstance(subkey, int):
                indices[i] = [subkey]
                squeezed_dims.append(self.dims[i])
            elif isinstance(subkey, list):
                indices[i] = subkey
            elif isinstance(subkey, slice):
                indices[i] = range(*subkey.indices(self.shape[i]))
                
        # can't squeeze all dimensions!!
        if len(squeezed_dims) == len(self.dims):
            raise IndexError("cannot return object with all remaining dimensions squeezed")
            
        # fail nicely if we fail at all
        try:
            toret = self.__finalize__(self.data[key], self.mask[key], indices)
            for dim in squeezed_dims:
                toret = toret.squeeze(dim)
            return toret
        except ValueError:
            raise IndexError("this type of slicing not implemented")
        
    def _get_index(self, dim, val, method=None):
        """
        Internal function to compute the bin index of the nearest 
        coordinate value to the input value
        """
        index = self.coords[dim]
        if method == 'nearest':
            i = (numpy.abs(index-val)).argmin()
        else:
            try:
                i = list(index).index(val)
            except Exception as e:
                args = (dim, str(e))
                msg = "error converting '%s' index; try setting `method = 'nearest'`: %s"
                raise IndexError(msg %args)
                
        return i
        
    def __reduce__(self):
        """
        The standard method for python pickle serialization
        """
        necessary = ['data', 'mask', 'dims', 'edges', 'coords', 'attrs']
        d = {k:getattr(self,k) for k in necessary}
        cls = self.__class__
        return (__unpickle__, (self.__class__, d, ))
        
    #--------------------------------------------------------------------------
    # user-called functions
    #--------------------------------------------------------------------------
    def copy(self):
        """
        Returns a copy of the DataSet
        """
        attrs = self.__copy_attrs__()
        cls = self.__class__
        return cls.__construct_direct__(self.data.copy(), self.mask.copy(), **attrs)
        
    def rename_variable(self, old_name, new_name):
        """
        Rename a variable in :attr:`data` from `old_name` to `new_name`
        
        Note that this procedure is performed in-place (does not 
        return a new DataSet)
        
        Parameters
        ----------
        old_name : str
            the name of the old varibale to rename
        new_name : str
            the desired new variable name
        
        Raises
        ------
        ValueError
            If `old_name` is not present in :attr:`variables`
        """
        import copy
        if old_name not in self.variables:
            raise ValueError("`%s` is not an existing variable name" %old_name)
        new_dtype = copy.deepcopy(self.data.dtype)
         
        names = list(new_dtype.names)
        names[names.index(old_name)] = new_name
        new_dtype.names = names
        self.data.dtype = new_dtype
        
    def sel(self, method=None, **indexers):
        """
        Return a new DataSet indexed by coordinate values along the 
        specified dimension(s) 
        
        Notes
        -----
        Scalar values used to index a specific dimension will result in 
        that dimension being squeezed. To keep a dimension of unit length,
        use a list to index (see examples below).
        
        Parameters
        ----------
        method : {None, 'nearest'}
            The method to use for inexact matches; if set to `None`, require
            an exact coordinate match, otherwise match the nearest coordinate
        **indexers : 
            the pairs of dimension name and coordinate value used to index
            the DataSet
            
        Returns
        -------
        sliced : DataSet
            a new DataSet holding the sliced data and coordinate grid
        
        Examples
        --------
        >>> pkmu
        <DataSet: dims: (k_cen: 200, mu_cen: 5), variables: ('mu', 'k', 'power')>
        
        >>> pkmu.sel(k_cen=0.4)
        <DataSet: dims: (mu_cen: 5), variables: ('mu', 'k', 'power')>
        
        >>> pkmu.sel(k_cen=[0.4])
        <DataSet: dims: (k_cen: 1, mu_cen: 5), variables: ('mu', 'k', 'power')>
        
        >>> pkmu.sel(k_cen=slice(0.1, 0.4), mu_cen=0.5)
        <DataSet: dims: (k_cen: 30), variables: ('mu', 'k', 'power')>
        """
        indices = [range(0, self.shape[i]) for i in range(len(self.dims))]
        squeezed_dims = []
        for dim in indexers:
            key = indexers[dim]
            i = self.dims.index(dim)
            
            if isinstance(key, list):
                indices[i] = [self._get_index(dim, k, method=method) for k in key]
            elif isinstance(key, slice):
                new_slice = []
                for name in ['start', 'stop']:
                    new_slice.append(self._get_index(dim, getattr(key, name), method=method))
                indices[i] = range(*slice(*new_slice).indices(self.shape[i]))
            elif not numpy.isscalar(key):
                raise IndexError("please index using a list, slice, or scalar value")
            else:
                indices[i] = [self._get_index(dim, key, method=method)]
                squeezed_dims.append(dim)
        
        # can't squeeze all dimensions!!
        if len(squeezed_dims) == len(self.dims):
            raise IndexError("cannot return object with all remaining dimensions squeezed")
        
        # check for empty slices
        for i, idx in enumerate(indices):
            if not len(idx):
                raise KeyError("trying to use empty slice for dimension '%s'" %self.dims[i])
        
        data = self.data.copy()
        mask = self.mask.copy()
        for i, idx in enumerate(indices):
            data = numpy.take(data, idx, axis=i)
            mask = numpy.take(mask, idx, axis=i)
        
        toret = self.__finalize__(data, mask, indices)
        for dim in squeezed_dims:
            toret = toret.squeeze(dim)
        return toret
                  
    @classmethod
    def from_nbkit(cls, d, meta):
        """
        Return a DataSet object from a dictionary of data and
        metadata
        
        Notes
        -----
        *   The dictionary `d` must also have entries for `dims`
            and `edges` which are used to construct the DataSet
        
        Parameters
        ----------
        d : dict 
            A dictionary holding the data variables, as well as the
            `dims` and `edges` values
        meta : dict
            dictionary of metadata to store in the :attr:`attrs` attribute
        
        Returns
        -------
        DataSet
            The newly constructed DataSet
        """
        d = d.copy() # copy so we don't edit for caller
        for name in ['dims', 'edges']:
            if name not in d:
                raise ValueError("must supply `%s` value in input dictionary" %name)
                
        edges = d.pop('edges')
        dims = d.pop('dims')
        return cls(dims, edges, d, **meta)
                
    def squeeze(self, dim=None):
        """
        Squeeze the DataSet along the specified dimension, which
        removes that dimension from the DataSet
        
        The behavior is similar to that of :func:`numpy.squeeze`.
        
        Parameters
        ----------
        dim : str, optional
            The name of the dimension to squeeze. If no dimension
            is provided, then the one dimension with unit length will 
            be squeezed
        
        Returns
        -------
        squeezed : DataSet
            a new DataSet instance, squeezed along one dimension
        
        Raises
        ------
        ValueError
            If the specified dimension does not have length one, or 
            no dimension is specified and multiple dimensions have 
            length one
        
        Examples
        --------
        >>> pkmu
        <DataSet: dims: (k_cen: 200, mu_cen: 1), variables: ('mu', 'k', 'power')>
        >>> pkmu.squeeze() # squeeze the mu dimension
        <DataSet: dims: (k_cen: 200), variables: ('mu', 'k', 'power')>
        """
        # infer the right dimension to squeeze
        if dim is None:
            dim = [k for k in self.dims if len(self.coords[k]) == 1]
            if not len(dim):
                raise ValueError("no available dimensions with length one to squeeze")
            if len(dim) > 1:
                raise ValueError("multiple dimensions available to squeeze -- please specify")
            dim = dim[0]
        else:
            if dim not in self.dims:
                raise ValueError("`%s` is not a valid dimension name" %dim)
            if len(self.coords[dim]) != 1:
                raise ValueError("the `%s` dimension must have length one to squeeze" %dim)
    
        # remove the dimension from the grid
        i = self.dims.index(dim)
        toret = self.copy()
        toret.dims.pop(i); toret.edges.pop(dim); toret.coords.pop(dim)
        if not len(toret.dims):
            raise ValueError("cannot squeeze the only remaining axis")
        
        # construct new object with squeezed data/mask
        attrs = toret.__copy_attrs__()
        d = toret.data.squeeze(axis=i)
        m = toret.mask.squeeze(axis=i)
        return self.__construct_direct__(d, m, **attrs)
        
    def average(self, dim, **kwargs):
        """
        Compute the average of each variable over the specified dimension.
        
        Parameters
        ----------
        dim : str
            The name of the dimension to average over
        **kwargs :
            Additional keywords to pass to :func:`DataSet.reindex`. See the 
            documentation for :func:`DataSet.reindex` for valid keywords.
        
        Returns
        -------
        averaged : DataSet
            A new DataSet, with data averaged along one dimension, 
            which reduces the number of dimension by one
        """
        spacing = (self.edges[dim][-1] - self.edges[dim][0])
        toret = self.reindex(dim, spacing, **kwargs)
        toret = toret.sel(**{dim:toret.coords[dim][0]})
        return toret.squeeze(dim)
        
    def reindex(self, 
                    dim, 
                    spacing, 
                    weights=None, 
                    force=True, 
                    return_spacing=False):
        """
        Reindex the dimension `dim` by averaging over multiple coordinate bins, 
        optionally weighting by `weights`. Return a new DataSet holding the 
        re-binned data
        
        Notes
        -----
        *   We can only re-bin to an integral factor of the current 
            dimension size in order to inaccuracies when re-binning to 
            overlapping bins
        *   Variables specified in :attr:`DataSet._fields_to_sum` will 
            be summed when re-indexing, instead of averaging
        
        
        Parameters
        ----------
        dim : str
            The name of the dimension to average over
        spacing : float
            The desired spacing for the re-binned data. If `force = True`,
            the spacing used will be the closest value to this value, such
            that the new bins are N times larger, when N is an integer
        weights : array_like or str, optional (`None`)
            An array to weight the data by before re-binning, or if
            a string is provided, the name of a data column to use as weights
        force : bool, optional
            If `True`, force the spacing to be a value such
            that the new bins are N times larger, when N is an integer, 
            otherwise, raise an exception. Default is `True`
        return_spacing : bool, optional
            If `True`, return the new spacing as the second return value. 
            Default is `False`.
            
        Returns
        -------
        rebinned : DataSet
            A new DataSet instance, which holds the rebinned coordinate
            grid and data variables
        spacing : float, optional
            If `return_spacing` is `True`, the new coordinate spacing 
            will be returned
        """        
        i = self.dims.index(dim)
        
        # determine the new binning
        old_spacings = numpy.diff(self.coords[dim])
        if not numpy.array_equal(old_spacings, old_spacings):
            raise ValueError("`reindex` requires even bin spacings")
        old_spacing = old_spacings[0]
        
        factor = numpy.round(spacing/old_spacing).astype('int')
        if not factor:
            raise ValueError("new spacing must be smaller than original spacing of %.2e" %old_spacing)
        if factor == 1:
            raise ValueError("closest binning size to input spacing is the same as current binning")
        if old_spacing*factor != spacing and not force: 
            raise ValueError("if `force = False`, new bin spacing must be an integral factor smaller than original")
        
        # make a copy of the data
        data = self.data.copy()
                
        # get the weights
        if isinstance(weights, str):
            if weights not in self.variables:
                raise ValueError("cannot weight by `%s`; no such column" %weights)
            weights = self.data[weights]
            
        edges = self.edges[dim]
        
        # check if we need to discard bins from the end
        leftover = self.shape[i] % factor
        if leftover and not force:
            args = (leftover, old_spacing*factor)
            raise ValueError("cannot re-bin because they are %d extra bins, using spacing = %.2e" %args)
        if leftover:
            sl = [slice(None, None)]*len(self.dims)
            sl[i] = slice(None, -leftover)
            data = data[sl]
            if weights is not None: weights = weights[sl]
            edges = edges[:-leftover]
            
        # new edges
        new_shape = list(self.shape)
        new_shape[i] /= factor
        new_edges = numpy.linspace(edges[0], edges[-1], new_shape[i]+1)
        
        # the re-binned data
        new_data = numpy.empty(new_shape, dtype=self.data.dtype)
        for name in self.variables:
            operation = numpy.nanmean
            weights_ = weights
            if weights is not None or name in self._fields_to_sum:
                operation = numpy.nansum
                if name in self._fields_to_sum: weights_ = None
            new_data[name] = bin_ndarray(data[name], new_shape, weights=weights_, operation=operation)
        
        # the new mask
        new_mask = numpy.zeros_like(new_data, dtype=bool)
        for name in self.variables:
            new_mask = numpy.logical_or(new_mask, ~numpy.isfinite(new_data[name]))
        
        # construct new object
        kw = self.__copy_attrs__()
        kw['edges'][dim] = new_edges
        kw['coords'][dim] = 0.5*(new_edges[1:] + new_edges[:-1])
        toret = self.__construct_direct__(new_data, new_mask, **kw)
        
        return (toret, spacing) if return_spacing else toret
        
def from_1d_measurement(dims, d, meta, columns=None, **kwargs):
    """
    Return a DataSet object from columns of data and
    any additional meta data
    """
    meta = meta.copy() # make a copy
    if columns is None: 
        columns = meta.pop('columns', None)
    
    if 'edges' not in meta:
        raise ValueError("must supply `edges` value in `meta` input")
    if columns is None:
        raise ValueError("if `columns` key not in `meta`, then `columns` keyword must not be `None`")
    
    d = {col : d[:,i] for i, col in enumerate(columns)}
    d['edges'] = [meta.pop('edges')]
    d['dims'] = dims
    meta.update(kwargs)
    return DataSet.from_nbkit(d, meta)
    
def from_2d_measurement(dims, d, meta, **kwargs):
    """
    Return a DataSet object from a dictionary of data 
    and any additional data
    """
    if 'edges' not in d:
        raise ValueError("must supply `edges` value in `d` input")

    d['dims'] = dims
    meta.update(kwargs)
    return DataSet.from_nbkit(d, meta)


class Power1dDataSet(DataSet):
    """
    A `DataSet` that holds a 1D power spectrum in bins of `k`
    """
    _fields_to_sum = ['modes']
    
    def __init__(self, edges, variables, **kwargs):
        super(Power1dDataSet, self).__init__(['k'], [edges], variables, **kwargs)
        for field in ['modes']:
            if field in self: self._fields_to_sum.append(field)
        
    @classmethod
    def from_nbkit(cls, d, meta, columns=None, **kwargs):
        """
        Return a `Power1dDataSet` instance taking the return values
        of `files.Read1DPlainText` as input
               
        Parameters
        ----------
        d : array_like
            the 1D data stacked vertically, such that each columns
            represents a separate data variable
        meta : dict
            any additional metadata to store as part of the 
            P(k) measurement
        columns : list
            list of the column names -- required if `columns`
            not in `meta` dictionary
        
        Examples
        --------
        >>> from nbodykit import files
        >>> power = Power1dDataSet.from_nbkit(*files.Read1DPlainText(filename))
        """
        toret = from_1d_measurement(['k'], d, meta, columns=columns, **kwargs)
        toret.__class__ = cls
        return toret
        
class Power2dDataSet(DataSet):
    """
    A `DataSet` that holds a 2D power spectrum in bins of `k` and `mu`
    """
    _fields_to_sum = ['modes']
    
    def __init__(self, edges, variables, **kwargs):
        super(Power2dDataSet, self).__init__(['k','mu'], edges, variables, **kwargs)
        
    @classmethod
    def from_nbkit(cls, d, meta, **kwargs):
        """
        Return a `Power2dDataSet` instance taking the return values
        of `files.Read2DPlainText` as input
               
        Parameters
        ----------
        d : dict
            dictionary holding the `edges` data, as well as the
            data columns for the 2D measurement
        meta : dict
            any additional metadata to store as part of the 
            2D measurement
        
        Examples
        --------
        >>> from nbodykit import files
        >>> power = Power2dDataSet.from_nbkit(*files.Read2DPlainText(filename))
        """
        toret = from_2d_measurement(['k', 'mu'], d, meta, **kwargs)
        toret.__class__ = cls
        return toret
        
class Corr1dDataSet(DataSet):
    """
    A `DataSet` that holds a 1D correlation function in bins of `r`
    """
    _fields_to_sum = ['N', 'RR']
    
    def __init__(self, edges, variables, **kwargs):
        
        super(Corr1dDataSet, self).__init__(['r'], [edges], variables, **kwargs)
        for field in ['N', 'RR']:
            if field in self: self._fields_to_sum.append(field)
            
    @classmethod
    def from_nbkit(cls, d, meta, columns=None, **kwargs):
        """
        Return a `Corr1dDataSet` instance taking the return values
        of `files.Read1DPlainText` as input
               
        Parameters
        ----------
        d : array_like
            the 1D data stacked vertically, such that each columns
            represents a separate data variable
        meta : dict
            any additional metadata to store as part of the 
            P(k) measurement
        columns : list
            list of the column names -- required if `columns`
            not in `meta` dictionary
        
        Examples
        --------
        >>> from nbodykit import files
        >>> corr = Power1dDataSet.from_nbkit(*files.Read1DPlainText(filename))
        """
        toret = from_1d_measurement(['r'], d, meta, columns=columns, **kwargs)
        toret.__class__ = cls
        return toret
        
class Corr2dDataSet(DataSet):
    """
    A `DataSet` that holds a 2D correlation in bins of `k` and `mu`
    """
    _fields_to_sum = ['N', 'RR']
    
    def __init__(self, edges, variables, **kwargs):
        super(Corr2dDataSet, self).__init__(['r','mu'], edges, variables, **kwargs)
        
    @classmethod
    def from_nbkit(cls, d, meta, **kwargs):
        """
        Return a `Corr2dDataSet` instance taking the return values
        of `files.Read2DPlainText` as input
               
        Parameters
        ----------
        d : dict
            dictionary holding the `edges` data, as well as the
            data columns for the 2D measurement
        meta : dict
            any additional metadata to store as part of the 
            2D measurement
        
        Examples
        --------
        >>> from nbodykit import files
        >>> corr = Corr2dDataSet.from_nbkit(*files.Read2DPlainText(filename))
        """
        toret = from_2d_measurement(['r', 'mu'], d, meta, **kwargs)
        toret.__class__ = cls
        return toret
    
        





from nbodykit.utils.config import autoassign, attribute
from astropy import cosmology, units

class CosmologyBase(object):
    """
    Base class for computing cosmology-dependent quantities, possibly
    sampling them at at set of points and returning interpolated results 
    (for speed purposes)
    """
    class sampled_function:
        """
        Class to represent a "sampled" version of a function
        """
        def __init__(self, func, x, *args, **kwargs):
            from scipy import interpolate
            self.func = func
            self.x = x

            # assign function name and docs
            self.__name__ = None
            self.__doc__  = None
            if func.__name__ != None:
                self.__name__ = self.func.__name__ + " [Sampled to %i pts]" %len(x)
            if func.__doc__ != None:
                self.__doc__ = "sampled function : \n\n" + self.func.__doc__

            self.spline = interpolate.InterpolatedUnivariateSpline(x, func(x,*args,**kwargs))

        def __call__(self, y):
            return self.spline(y)
            
    def sample(self, methodname, x, *args, **kwargs):
        if not hasattr(self, methodname):
            raise ValueError("no such method '%s' to sample" %methodname)

        # unsample first
        if self.is_sampled(methodname):
            self.unsample(methodname)
        
        # set the sampled function
        tmp = getattr(self, methodname)
        setattr(self, methodname, self.sampled_function(tmp, x, *args, **kwargs))

    def unsample(self,methodname):
        if self.is_sampled(methodname):
            tmp = getattr(self, methodname).func
            setattr(self, methodname, tmp)
        else:
            raise ValueError("cannot unsample %s" %methodname)
        
    def is_sampled(self,methodname):
        return getattr(self,methodname).__class__ == self.sampled_function
                
def neutrino_mass(value):
    """
    Function to cast an input string or list to a `astropy.units.Quantity`,
    with units of `eV` to represent neutrino mass
    """    
    if isinstance(value, str):
        value = [float(i) for i in value.split()]
    
    if isinstance(value, list):
        if len(value) == 1:
            value = value[0]
        elif len(value) != 3:
            raise ValueError("either a single neutrino mass, or a mass for each of the 3 species must be provided")
    return value
    
class Cosmology(CosmologyBase):
    """
    A class for computing cosmology-dependent quantites, which uses
    `astropy.Cosmology` to do the calculations
    
    Notes
    -----
    *   this class supports the `wCDM` and `FlatwCDM` classes from `astropy`,
    *   additions to the fiducial `LCDM` model include the dark energy equation 
        of state `w0` and massive neutrinos
    *   if `flat = True`, the dark energy density is set automatically
    *   the underlying astropy class is stored as the `engine` attribute
    """  
    
    @autoassign
    @attribute('flat', type=bool, help="if `True`, automatically set `Ode0` such that `Ok0` is zero")
    @attribute('m_nu', type=neutrino_mass, help="mass of neutrino species in eV")
    @attribute('Neff', type=float, help="effective number of neutrino species")
    @attribute('Tcmb0', type=float, help="temperature of the CMB in K at z=0")
    @attribute('w0', type=float, help="dark energy equation of state")
    @attribute('Ode0', type=float, help="dark energy density/critical density at z=0")
    @attribute('Om0', type=float, help="matter density/critical density at z=0")
    @attribute('H0', type=float, help="the Hubble constant at z=0, in km/s/Mpc")
    def __init__(self, H0=67.6, Om0=0.31, Ode0=0.69, w0=-1., Tcmb0=2.7255, 
                    Neff=3.04, m_nu=0., flat=False):

        # convert neutrino mass to a astropy `Quantity`
        self.m_nu = units.Quantity(self.m_nu, 'eV')
        
        # initialize the cosmology object
        kw = {k:getattr(self,k) for k in ['w0', 'Tcmb0', 'Neff', 'm_nu']}
        if not self.flat:
            self.engine = cosmology.wCDM(self.H0, self.Om0, self.Ode0, **kw)
        else:
            self.engine = cosmology.FlatwCDM(self.H0, self.Om0, **kw)
            self.Ode0 = self.engine.Ode0 # set this automatically
    
    def comoving_distance(self, z):
        """
        Returns the comoving distance to z in units of `Mpc/h`
        """
        return self.engine.comoving_distance(z).value * self.engine.h



from nbodykit.extensionpoints import Algorithm, DataSource

import numpy
import logging

class Describe(Algorithm):
    plugin_name = "Describe"
    logger = logging.getLogger(plugin_name)

    def __init__(self, datasource, column='Position'):
        pass

    @classmethod
    def register(cls):
        s = cls.schema
        s.description = "describe a specific column of the data source"
        s.add_argument("datasource", type=DataSource.from_config, 
            help='the DataSource plugin; run `nbkit.py --list-datasources` for all options')
        s.add_argument("column", type=str, help='the column to describe')
     
    def finalize_attributes(self):
        pass

    def run(self):
        """
        Run the algorithm, which does nothing
        """
        left = []
        right = []
        with self.datasource.open() as stream:
            for [pos] in stream.read([self.column]):
                left.append(numpy.min(pos, axis=0))
                right.append(numpy.max(pos, axis=0))
            left = numpy.min(left, axis=0)
            right = numpy.max(right, axis=0)
            left = numpy.min(self.comm.allgather(left), axis=0)
            right = numpy.max(self.comm.allgather(right), axis=0)
        return left, right

    def save(self, output, data):
        left, right = data
        if self.comm.rank == 0:
            template = "DataSource %s Column %s : min = %s max = %s\n"
            if output == '-' or output is None:
                import sys
                output = sys.stdout
            else:
                output = file(output, 'w')
            output.write(template % 
                (self.datasource, self.column, str(left), str(right)))




from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir, bin_dir

class RunBatchAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'batch')
           

def bash_fixture(cls, name, **kws):
    """
    Return a run fixture that uses bash to execute the param file
    """
    param_file = os.path.join(cls.run_dir, cls.param_file)
    cmd = "bash %s" %param_file
    return cls.class_fixture(cmd=cmd, stdout='run.stdout', stderr="run.stderr", **kws)
    
def batch_fixture(cls, name, **kws):
    """
    Return a run fixture that runs nbkit-batch
    """
    param_file = os.path.join(cls.run_dir, cls.param_file)
    extra_file = os.path.join(cls.run_dir, cls.extra_file)

    args = (bin_dir, name, param_file, extra_file)
    cmd = "mpirun -np 6 python %s/nbkit-batch.py %s 2 -c %s -i \"los: [x, y, z]\" --extras %s --debug" %args
    return cls.class_fixture(cmd=cmd, stdout='run.stdout', stderr="run.stderr", **kws)

@add_run_fixture(__name__, RunBatchAlgorithm, 'FFTPower', timeout=30, make_fixture=bash_fixture)
class TestStdin(unittest.TestCase):
    param_file  = "test_stdin.sh"
    output_file = "test_stdin.dat"
    datasources = ['fastpm_1.0000']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')
        
@add_run_fixture(__name__, RunBatchAlgorithm, 'FFTPower', timeout=120, make_fixture=batch_fixture)
class TestBatch(unittest.TestCase):
    param_file  = "test_power_batch.template"
    extra_file  = "extra.template"
    output_file = "test_stdin.dat"
    datasources = ['fastpm_1.0000']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')




from nbodykit.extensionpoints import Painter
import numpy
import logging

logger = logging.getLogger('DefaultPainter')

class DefaultPainter(Painter):
    plugin_name = "DefaultPainter"

    def __init__(self, weight=None):
        pass
        
    @classmethod
    def register(cls):
        s = cls.schema
        s.add_argument("weight", help="the column giving the weight for each object")

    def paint(self, pm, datasource):
        """
        Paint the ``DataSource`` specified by ``input`` onto the 
        ``ParticleMesh`` specified by ``pm``
    
        Parameters
        ----------
        pm : ``ParticleMesh``
            particle mesh object that does the painting
        datasource : ``DataSource``
            the data source object representing the field to paint onto the mesh
            
        Returns
        -------
        stats : dict
            dictionary of statistics, usually only containing `Ntot`
        """
        pm.real[:] = 0
        stats = {}
        
        # open the datasource stream (with no defaults)
        with datasource.open() as stream:

            Nlocal = 0
            if self.weight is None:
                for [position] in stream.read(['Position']):
                    Nlocal += self.basepaint(pm, position)
            else:
                for position, weight in stream.read(['Position', self.weight]):
                    Nlocal += self.basepaint(pm, position, weight)
        
        stats['Ntot'] = self.comm.allreduce(Nlocal)
        return stats
            



from nbodykit.extensionpoints import DataSource
from nbodykit.utils import selectionlanguage
import numpy
import logging

logger = logging.getLogger('PlainText')

class PlainTextDataSource(DataSource):
    """
    DataSource to read a plaintext file, using `numpy.recfromtxt`
    to do the reading
    
    Notes
    -----
    *   data file is assumed to be space-separated
    *   commented lines must begin with `#`, with all other lines
        providing data values to be read
    *   `names` parameter must be equal to the number of data
        columns, otherwise behavior is undefined
    """
    plugin_name = "PlainText"
    
    def __init__(self, path, names, BoxSize, 
                    usecols=None, poscols=['x','y','z'], velcols=None, 
                    rsd=None, posf=1., velf=1., select=None):     
        pass
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read data from a plaintext file using numpy"
        
        s.add_argument("path", type=str,
            help="the file path to load the data from")
        s.add_argument("names", type=str, nargs='*', 
            help="names of columns in text file or name of the data group in hdf5 file")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("usecols", type=str, nargs='*',
            help="only read these columns from file")
        s.add_argument("poscols", type=str, nargs=3,
            help="names of the position columns")
        s.add_argument("velcols", type=str, nargs=3,
            help="names of the velocity columns")
        s.add_argument("rsd", type=str, choices="xyz", 
            help="direction to do redshift distortion")
        s.add_argument("posf", type=float, 
            help="factor to scale the positions")
        s.add_argument("velf", type=float, 
            help="factor to scale the velocities")
        s.add_argument("select", type=selectionlanguage.Query, 
            help='row selection based on conditions specified as string, i.e., "Mass > 1e14"')
        
    def readall(self):
        """
        Read all available data, returning a dictionary
        
        This provides ``Position`` and optionally ``Velocity`` columns
        """
        
        # read in the plain text file as a recarray
        kwargs = {}
        kwargs['comments'] = '#'
        kwargs['names'] = self.names
        kwargs['usecols'] = self.usecols
        data = numpy.recfromtxt(self.path, **kwargs)
        nobj = len(data)
        
        # select based on input conditions
        if self.select is not None:
            mask = self.select.get_mask(data)
            data = data[mask]
        logger.info("total number of objects selected is %d / %d" % (len(data), nobj))
        
        toret = {}
        
        # get position and velocity, if we have it
        pos = numpy.vstack(data[k] for k in self.poscols).T.astype('f4')
        pos *= self.posf
        if self.velcols is not None:
            vel = numpy.vstack(data[k] for k in self.velcols).T.astype('f4')
            vel *= self.velf
            toret['Velocity'] = Velocity

        if self.rsd is not None:
            dir = "xyz".index(self.rsd)
            pos[:, dir] += vel[:, dir]
            pos[:, dir] %= self.BoxSize[dir]
        toret['Position'] = pos
        
        return toret




from nbodykit.extensionpoints import DataSource
import numpy
import logging
import bigfile

logger = logging.getLogger('FastPM')

class FastPMDataSource(DataSource):
    """
    DataSource to read snapshot files of the FastPM simulation
    """
    plugin_name = "FastPM"
    
    def __init__(self, path, BoxSize=None, bunchsize=4*1024*1024, rsd=None):
        
        BoxSize = numpy.empty(3, dtype='f8')
        f = bigfile.BigFileMPI(self.comm, self.path)
        header = f['header']
        BoxSize[:] = header.attrs['BoxSize'][0]
        OmegaM = header.attrs['OmegaM'][0]
        self.M0 = 27.75e10 * OmegaM * BoxSize[0] ** 3 / f['Position'].size
        self.size = f['Position'].size

        if self.comm.rank == 0:
            logger.info("File has boxsize of %s Mpc/h" % str(BoxSize))
            logger.info("Mass of a particle is %g Msun/h" % self.M0)

        if self.BoxSize is None:
            self.BoxSize = BoxSize
        else:
            if self.comm.rank == 0:
                logger.info("Overriding boxsize as %s" % str(self.BoxSize))
        
    
    @classmethod
    def register(cls):
        """
        Fill the attribute schema associated with this class
        """
        s = cls.schema
        s.description = "read snapshot files of the FastPM simulation"
        
        s.add_argument("path", type=str,
            help="the file path to load the data from")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="override the size of the box; can be a scalar or a three vector")
        s.add_argument("rsd", type=str, choices="xyz", 
            help="direction to do redshift distortion")
        s.add_argument("bunchsize", type=int, 
            help="number of particles to read per rank in a bunch")
                
    def parallel_read(self, columns, full=False):
        f = bigfile.BigFileMPI(self.comm, self.path)
        header = f['header']
        boxsize = header.attrs['BoxSize'][0]
        RSD = header.attrs['RSDFactor'][0]
        if boxsize != self.BoxSize[0]:
            raise ValueError("Box size mismatch, expecting %g" % boxsize)

        readcolumns = set(columns)
        if self.rsd is not None:
            readcolumns = set(columns + ['Velocity'])
        if 'InitialPosition' in columns:
            readcolumns.add('ID')
            readcolumns.remove('InitialPosition')

        if 'Mass' in readcolumns: 
            readcolumns.remove('Mass')
            
        # remove columns not in the file (None will be returned)
        for col in list(readcolumns):
            if col not in f:
                readcolumns.remove(col)
            
        done = False
        i = 0
        while not numpy.all(self.comm.allgather(done)):
            ret = []
            dataset = bigfile.BigData(f, readcolumns)

            Ntot = dataset.size
            start = self.comm.rank * Ntot // self.comm.size
            end = (self.comm.rank + 1) * Ntot // self.comm.size

            if not full:
                bunchstart = start + i * self.bunchsize
                bunchend = start + (i + 1) * self.bunchsize
                if bunchend > end: bunchend = end
                if bunchstart > end: bunchstart = end
            else:
                bunchstart = start
                bunchend = end

            if bunchend == end:
                done = True

            P = {}

            for column in readcolumns:
                data = dataset[column][bunchstart:bunchend]
                P[column] = data

            if 'Velocity' in P:
                P['Velocity'] *= RSD

            if 'Mass' in columns:
                P['Mass'] = numpy.ones(bunchend - bunchstart, dtype='u1') * self.M0

            if self.rsd is not None:
                dir = "xyz".index(self.rsd)
                P['Position'][:, dir] += P['Velocity'][:, dir]
                P['Position'][:, dir] %= self.BoxSize[dir]
            if 'InitialPosition' in columns:
                P['InitialPosition'] = numpy.empty((len(P['ID']), 3), 'f4')
                nc = int(self.size ** (1. / 3) + 0.5)
                id = P['ID'].copy()
                for nc in range(nc - 10, nc + 10):
                    if nc ** 3 == self.size: break
                for d in [2, 1, 0]:
                    P['InitialPosition'][:, d] = id % nc
                    id[:] //= nc
                cellsize = self.BoxSize[0] / nc
                P['InitialPosition'][:] += 0.5
                P['InitialPosition'][:] *= cellsize

            i = i + 1
            yield [P.get(column, None) for column in columns]



import os
import sys
import unittest
import pytest

# idea borrowed from xarray
PY3 = sys.version_info[0] >= 3
if PY3:
    from urllib.request import urlretrieve as _urlretrieve
else:
    from urllib import urlretrieve as _urlretrieve
    
from nbodykit import examples_dir
cache_dir = os.path.join('~', '.nbodykit')

def download_data(github_url, cache_dir):
    """
    Download the github url tarball to the cache directory
    """
    import tarfile
    
    # make the cache dir if it doesnt exist
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    
    # download the tarball locally
    tarball_link = os.path.join(github_url, 'tarball', 'master')
    tarball_local = os.path.join(cache_dir, 'master.tar.gz')
    _urlretrieve(tarball_link, tarball_local)
    
    # extract the tarball to the cache dir
    with tarfile.open(tarball_local) as tar:
    
        members = tar.getmembers()
        topdir = members[0].name
    
        for m in members[1:]:
            name = os.path.relpath(m.name, topdir)
            m.name = name
            tar.extract(m, path=cache_dir)
            
    # remove the downloaded tarball file
    if os.path.exists(tarball_local):
        os.remove(tarball_local)


def verify_data_in_cache(name, cache_dir=cache_dir,
                            github_url='https://github.com/bccp/nbodykit-data'):
    """
    Load a dataset from the online repository (requires internet).

    If a local copy is found then always use that to avoid network traffic.

    Parameters
    ----------
    name : str
        Name of the netcdf file containing the dataset
        ie. 'air_temperature'
    cache_dir : string, optional
        The directory in which to search for and write cached data.
    github_url : string
        Github repository where the data is stored
    """
    cache_dir = os.path.expanduser(cache_dir)
    longdir = os.path.join(cache_dir, 'data')
    localfile = os.path.join(longdir, name)

    # download and cache the nbodykit-data directory
    if not os.path.exists(localfile):
        download_data(github_url, cache_dir)
        
        # crash, if still no data
        if not os.path.exists(localfile):
            raise ValueError("filename `%s` does not exist in nbodykit-data directory" %name)



import logging

import h5py

from nbodykit import files
from nbodykit import halos

import numpy
import mpsort
from nbodykit.extensionpoints import Algorithm, DataSource

class TraceHaloAlgorithm(Algorithm):
    plugin_name = "TraceHalo"
    
    def __init__(self, dest, source, sourcelabel):
        pass
    
    @classmethod
    def register(cls):
        s = cls.schema
        s.description = " Calculate the halo property based on a different set of halo labels."

        s.add_argument("dest", type=DataSource.from_config, help="type: DataSource")
        s.add_argument("source", type=DataSource.from_config, help="type: DataSource")
        s.add_argument("sourcelabel", type=DataSource.from_config,
            help='DataSource of the source halo label files, the Label column is used.')

    def run(self):
        comm = self.comm

        with self.source.open() as source:
            [[ID]] = source.read(['ID'], full=True)

        Ntot = self.comm.allreduce(len(ID))

        start = sum(comm.allgather(len(ID))[:comm.rank])
        end   = sum(comm.allgather(len(ID))[:comm.rank+1])
        data = numpy.empty(end - start, dtype=[
                    ('Label', ('i4')), 
                    ('ID', ('i8')), 
                    ])
        data['ID'] = ID
        del ID
        with self.sourcelabel.open() as sourcelabel:
            [[data['Label'][...]]] = sourcelabel.read(['Label'], full=True)

        mpsort.sort(data, orderby='ID')

        label = data['Label'].copy()
        del data

        data = numpy.empty(end - start, dtype=[
                    ('ID', ('i8')), 
                    ('Position', ('f4', 3)), 
                    ('Velocity', ('f4', 3)), 
                    ])
        with self.dest.open() as dest:
            [[data['Position'][...]]] = dest.read(['Position'], full=True)
            [[data['Velocity'][...]]] = dest.read(['Velocity'], full=True)
            [[data['ID'][...]]] = dest.read(['ID'], full=True)
        mpsort.sort(data, orderby='ID')

        data['Position'] /= self.dest.BoxSize
        data['Velocity'] /= self.dest.BoxSize
        
        N = halos.count(label)
        hpos = halos.centerofmass(label, data['Position'], boxsize=1.0)
        hvel = halos.centerofmass(label, data['Velocity'], boxsize=None)
        return hpos, hvel, N, Ntot

    def save(self, output, data): 
        hpos, hvel, N, Ntot = data
        if self.comm.rank == 0:
            logging.info("Total number of halos: %d" % len(N))
            logging.info("N %s" % str(N))
            N[0] = 0
            with h5py.File(output, 'w') as ff:
                data = numpy.empty(shape=(len(N),), 
                    dtype=[
                    ('Position', ('f4', 3)),
                    ('Velocity', ('f4', 3)),
                    ('Length', 'i4')])
                
                data['Position'] = hpos
                data['Velocity'] = hvel
                data['Length'] = N
                
                # do not create dataset then fill because of
                # https://github.com/h5py/h5py/pull/606

                dataset = ff.create_dataset(
                    name='TracedFOFGroups', data=data
                    )
                dataset.attrs['Ntot'] = Ntot
                dataset.attrs['BoxSize'] = self.source.BoxSize
                dataset.attrs['source'] = self.source.string
                dataset.attrs['sourcelabel'] = self.sourcelabel.string
                dataset.attrs['dest'] = self.dest.string

            logging.info("Written %s" % output)




from pytest_pipeline import PipelineRun, mark
from . import os, sys, pytest
from . import verify_data_in_cache, cache_dir
from .. import bin_dir, pkg_dir

# global counter for the number of pipeline tests
_pipeline_tests = 0

def default_run_fixture(cls, name, timeout=60):
    """
    Create a class fixture to run the analysis pipeline
    
    Notes
    -----
    *   The input `cls` should have attributes `param_file` which specifies
        the input configuration file and `run_dir`, which specifies where 
        `param_file` lives
    *   The analysis run will be performed in parallel using 2 processes
    
    Parameters
    ----------
    cls : subclass of PipelineRun
        a subclass of PipelineRun that will create the run fixture
    name : str
        the name of the nbodykit Algorithm to run
    timeout : int, optional
        an optional timeout in seconds -- if this limit is exceeded, the
        analysis run is terminated and the test fails
    """
    param_file = os.path.join(cls.run_dir, cls.param_file)    
    cmd = "mpirun -n 2 python %s/nbkit.py %s %s" %(bin_dir, name, param_file)
    return cls.class_fixture(cmd=cmd, stdout='run.stdout', stderr="run.stderr", timeout=timeout)
    
def add_run_fixture(modname, basecls, name, make_fixture=default_run_fixture, timeout=60):
    """
    A class decorator that creates a run fixture and adds it to the class as 
    a ``pytest`` fixture 
    
    Parameters
    ----------
    modname : str
        the name of the module to which the fixture class should be attached; this
        is needed so that each test module knows where to find the relevant fixtures
    basecls : subclass of PipelineRun
        the base class for the run fixture, with any functions added that will be
        executed prior to the pipeline run
    name : str
        the name of the nbodykit Algorithm to run
    make_fixture : callable, optional
        a function that returns a class fixture that runs the pipeline; default is
        ``default_run_fixture``
    timeout : int, optional
        an optional timeout in seconds -- if this limit is exceeded, the
        analysis run is terminated and the test fails; default is 60 seconds
    """
    global _pipeline_tests
        
    def class_rebuilder(cls):        
        # create the run fixture class and attach the relevant info
        new_cls = type("RunClass%d" %_pipeline_tests, (basecls, cls), {})
    
        # make the run fixture and attach it to the module
        run_name = 'run_%s_%d' %(modname, _pipeline_tests)
        mod = sys.modules[modname]
        setattr(mod, run_name, make_fixture(new_cls, name, timeout=timeout))
    
        # initialize the mark decorator
        decorator = pytest.mark.usefixtures(run_name)
        
        return decorator(cls)
    
    _pipeline_tests += 1
    return class_rebuilder

class RunAlgorithm(PipelineRun):
    """
    Run a pipeline that tests an ``Algorithm``. Prior to running
    this class will: 
    
        * load the necessary data to the cache directory
        * initialize the ``output`` directory in ``nbodykit/examples``
        * set the ``NBKIT_CACHE`` and ``NBKIT_HOME`` directories
    """
    @mark.before_run
    def verify_cache(self):
        """
        Verify that all the necessary data has been downloaded
        and cached in order to run this test
        """
        for d in self.datasources:
            verify_data_in_cache(d)
            
    @mark.before_run
    def initialize_output(self):
        """
        Make the ``output`` directory in ``nbodykit/examples`, if 
        it does not exist yet
        """
        results_dir = os.path.join(pkg_dir, 'examples', 'output')
        if not os.path.exists(results_dir):
            os.makedirs(results_dir)
            
    @mark.before_run
    def set_env(self):
        """
        Set the ``NBKIT_CACHE`` and ``NBKTIT_HOME`` environment
        variables
        """
        if 'NBKIT_CACHE' not in os.environ:
            os.environ['NBKIT_CACHE'] = os.path.expanduser(cache_dir)
        if 'NBKIT_HOME' not in os.environ:
            os.environ['NBKIT_HOME'] = pkg_dir


from nbodykit.extensionpoints import Algorithm, DataSource
import numpy
import os

def binning_type(s):
    """
    Type conversion for use on the command-line that converts 
    a string to an array of bin edges
    """
    if os.path.isfile(s):
        return numpy.loadtxt(s)
    else:
        supported = ["`linspace: min max Nbins`", "`logspace: logmin logmax Nbins`"]
        try:
            f, params = s.split(':')
            params = list(map(float, params.split()))
            params[-1] = int(params[-1]) + 1

            if not hasattr(numpy, f): raise Exception
            if len(params) != 3: raise Exception

            return getattr(numpy, f)(*params)
        except:
            raise TypeError("supported binning format: [ %s ]" %", ".join(supported))


class PairCountCorrelationAlgorithm(Algorithm):
    """
    Algorithm to compute the 1d or 2d correlation function and multipoles
    via direct pair counting
    """
    plugin_name = "PairCountCorrelation"

    def __init__(self, mode, rbins, field, other=None, subsample=1, 
                    los='z', Nmu=10, poles=[]):
                    
        # construct the input fields list
        self.inputs = [self.field]
        if self.other is not None:
            self.inputs.append(self.other)
        
    @classmethod
    def register(cls):  
        s = cls.schema
        s.description = "correlation function calculator via pair counting"
    
        # the positional arguments
        s.add_argument("mode", type=str, choices=["1d", "2d"],
            help='measure the correlation function in `1d` or `2d`') 
        s.add_argument("rbins", type=binning_type, 
            help='the string specifying the binning to use') 
        s.add_argument("field", type=DataSource.from_config, 
            help='the first `DataSource` of objects to correlate; '
                 'run `nbkit.py --list-datasources` for all options')
        s.add_argument("other", type=DataSource.from_config, 
            help='the other `DataSource` of objects to cross-correlate with; '
                 'run `nbkit.py --list-datasources` for all options')
        s.add_argument("subsample", type=int, help='use 1 out of every N points')
        s.add_argument("los", choices="xyz",
            help="the line-of-sight: the angle `mu` is defined with respect to")
        s.add_argument("Nmu", type=int,
            help='if `mode == 2d`, the number of mu bins covering mu=[-1,1]')
        s.add_argument('poles', nargs='*', type=int,
            help='compute the multipoles for these `ell` values from xi(r,mu)')
   
    def run(self):
        """
        Run the pair-count correlation function and return the result
        
        Returns
        -------
        edges : list or array_like
            the array of 1d bin edges or a list of the bin edges in each dimension
        result : dict
            dictionary holding the data results (with associated names as keys) --
            this results `corr`, `RR`, `N` + the mean bin values
        """
        from nbodykit import measurestats
    
        # check multipoles parameters
        if len(self.poles) and self.mode == '2d':
            raise ValueError("you specified multipole numbers but `mode` is `2d` -- perhaps you meant `1d`")

        # set Nmu to 1 if doing 1d
        if self.mode == "1d": self.Nmu = 0

        # do the work
        kw = {'comm':self.comm, 'subsample':self.subsample, 'Nmu':self.Nmu, 'los':self.los, 'poles':self.poles}
        pc, xi, RR = measurestats.compute_brutal_corr(self.inputs, self.rbins, **kw)

        # format the results
        if self.mode == '1d':
            if len(self.poles):
                cols = ['r'] + ['corr_%d' %l for l in self.poles] + ['RR', 'N']
                result = [pc.mean_centers] + [xi[:,i] for i in range(len(self.poles))] + [RR, pc.pair_counts]
            else:
                cols = ['r', 'corr', 'RR', 'N']
                result = [pc.mean_centers, xi, RR, pc.pair_counts]
        else:
            cols = ['r', 'mu', 'corr', 'RR', 'N']
            r, mu = pc.mean_centers
            result = [r, mu, xi, RR, pc.pair_counts]

        return pc.edges, dict(zip(cols, result))
        
    def save(self, output, result):
        """
        Save the result returned by `run()` to the filename specified by `output`
        
        Parameters
        ----------
        output : str
            the string specifying the file to save
        result : tuple
            the tuple returned by `run()` -- first argument specifies the bin
            edges and the second is a dictionary holding the data results
        """
        from nbodykit.storage import MeasurementStorage
        
        # only master writes
        if self.comm.rank == 0:
            
            edges, result = result
            storage = MeasurementStorage.create(self.mode, output)
        
            cols = list(result.keys())
            values = list(result.values())
            storage.write(edges, cols, values)

            






import numpy

def make_gaussian_fields(pm, isotropic_power, random, compute_velocity=True):
    """
    Make a Gaussian realization of a density field, and if desired, the
    corresponding linear velocity fields
    """
    # clear the mesh
    pm.clear()
            
    # assign Gaussian rvs with mean 0 and unit variance
    pm.real[:] = random.normal(size=pm.real.shape)
    
    # initialize the velocity arrays
    if compute_velocity:
        vel_k = numpy.repeat((numpy.ones_like(pm.complex)*1j)[None], 3, axis=0)
        vel_x = numpy.repeat(numpy.zeros_like(pm.real)[None], 3, axis=0)
        
    # FT to k-space
    pm.r2c()
    
    # normalization
    norm = pm.Nmesh**3 / pm.BoxSize.prod()
    
    # loop over yz planes
    for islab in range(len(pm.k[0])):
        
        # now kslab stores k**2
        kslab = numpy.float64(pm.k[0][islab] ** 2)
        for ki in pm.k[1:]:
            kslab = kslab + ki[0]**2

        if len(kslab.flat) == 0:
            continue
    
        # the isotropic power (function of k)
        power = isotropic_power((kslab**0.5).flatten())
            
        # multiply complex field by sqrt of power
        pm.complex[islab,...].flat *= (power*norm)**0.5
        
        # set k == 0 to zero (zero x-space mean)
        zero_idx = kslab == 0. 
        pm.complex[islab, zero_idx] = 0.
        
        # do the velocity
        if compute_velocity:
            with numpy.errstate(invalid='ignore'):
                vel_k[0, islab, ...] *= pm.k[0][islab] / kslab * pm.complex[islab]
                vel_k[1, islab, ...] *= pm.k[1][0] / kslab * pm.complex[islab]
                vel_k[2, islab, ...] *= pm.k[2][0] / kslab * pm.complex[islab] 
                
            # no bulk velocity
            for i in range(3):    
                vel_k[i, islab, zero_idx] = 0.       
        
    # FT to x-space
    pm.c2r()
    delta = pm.real.copy()
    
    if not compute_velocity:
        return delta
    else:
        for i in range(3):
            pm.complex[:] = vel_k[i][:]
            pm.c2r()
            vel_x[i] = pm.real[:]
    
        return delta, vel_x
    

    
def lognormal_transform(density):
    """
    Apply a lognormal transformation to the density
    """
    toret = numpy.exp(density)
    toret /= toret.mean()
    return toret
    
    

def poisson_sample_mesh(pm, nbar, density, velocity=None):
    """
    Poisson sample the mesh
    """
    # mean number of objects per cell
    overallmean = pm.BoxSize.prod() / pm.Nmesh**3 * nbar
    
    # number of objects in each cell
    cellmean = density*overallmean
    
    # number of objects in each cell
    N = numpy.random.poisson(cellmean)
    Ntot = N.sum()
    pts = N.nonzero() # indices of nonzero points
    
    # setup the coordinate grid
    x = numpy.squeeze(pm.x[0])[pts[0]]
    y = numpy.squeeze(pm.x[1])[pts[1]]
    z = numpy.squeeze(pm.x[2])[pts[2]]
    
    # coordinates of each object (placed randomly in each cell)
    H = pm.BoxSize / pm.Nmesh
    x = numpy.repeat(x, N[pts]) + numpy.random.uniform(0, H[0], size=Ntot)
    y = numpy.repeat(y, N[pts]) + numpy.random.uniform(0, H[1], size=Ntot)
    z = numpy.repeat(z, N[pts]) + numpy.random.uniform(0, H[2], size=Ntot)
    
    # enforce periodic 
    x %= pm.BoxSize[0]
    y %= pm.BoxSize[1]
    z %= pm.BoxSize[2]
    pos = numpy.vstack([x, y, z]).T
    
    if velocity is not None:
        vx = numpy.repeat(velocity[0][pts], N[pts])
        vy = numpy.repeat(velocity[1][pts], N[pts])
        vz = numpy.repeat(velocity[2][pts], N[pts])
        vel = numpy.vstack([vx, vy, vz]).T
        
        return pos, vel
    else:
        return pos
    
    

        
                

        
    
    
    


from setuptools import setup, find_packages
from glob import glob
setup(name="nbodykit", version="0.1.1",
      author="Yu Feng, Nick Hand, et al",
      maintainer="Yu Feng",
      maintainer_email="rainwoodman@gmail.com",
      description="Data analysis of cosmology simulations in parallel.",
      url="http://github.com/bccp/nbodykit",
      zip_safe=False,
      package_dir = {'nbodykit': 'nbodykit'},
      package_data = {'nbodykit': list(glob('nbodykit/plugins/*/*.py'))},
      include_package_data=True,
      packages = find_packages(),
      install_requires=['numpy'],
      requires=['sharedmem', 'pmesh', 'pfft', 'kdcount', 'mpsort', 'scipy', 'bigfile'],
      scripts=['bin/nbkit.py', 'bin/nbkit-batch.py'],
)




from nbodykit.extensionpoints import Transfer
import numpy

class AnisotropicCIC(Transfer):
    """
    Divide by a kernel in Fourier space to account for
    the convolution of the gridded quantity with the 
    CIC window function in configuration space
    """
    plugin_name = "AnisotropicCIC"
    
    def __init__(self):
        pass
    
    @classmethod
    def register(cls):
        pass
        
    def __call__(self, pm, complex):
        for wi in pm.w:
            tmp = (1 - 2. / 3 * numpy.sin(0.5 * wi) ** 2) ** 0.5
            complex[:] /= tmp
            
class AnisotropicTSC(Transfer):
    """
    Divide by a kernel in Fourier space to account for
    the convolution of the gridded quantity with the 
    TSC window function in configuration space
    """
    plugin_name = "AnisotropicTSC"
    
    def __init__(self):
        pass
    
    @classmethod
    def register(cls):
        pass
        
    def __call__(self, pm, complex):
        for wi in pm.w:
            s = numpy.sin(0.5 * wi)**2
            tmp = (1 - s + 2./15 * s**2) ** 0.5
            complex[:] /= tmp
            
            
class NormalizeDC(Transfer):
    """
    Removes the DC amplitude in Fourier space, which effectively
    divides by the mean in configuration space
    """
    plugin_name = "NormalizeDC"
    def __init__(self):
        pass
    
    @classmethod
    def register(cls):
        pass

    def __call__(self, pm, complex):
        ind = []
        value = 0.
        found = True
        for wi in pm.w:
            if (wi != 0).all():
                found = False
                break
            ind.append((wi == 0).nonzero()[0][0])
        if found:
            ind = tuple(ind)
            value = numpy.abs(complex[ind])
        value = pm.comm.allreduce(value)
        complex[:] /= value
        
class RemoveDC(Transfer):
    """
    Remove the DC amplitude, which sets the mean of the 
    field in configuration space to zero
    """
    plugin_name = "RemoveDC"
    def __init__(self):
        pass

    @classmethod
    def register(cls):
        pass

    def __call__(self, pm, complex):
        ind = []
        for wi in pm.w:
            if (wi != 0).all():
                return
            ind.append((wi == 0).nonzero()[0][0])
        ind = tuple(ind)
        complex[ind] = 0.


from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunTidalTensorAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'tidaltensor')
           

@add_run_fixture(__name__, RunTidalTensorAlgorithm, 'TidalTensor')
class TestTPMSnapshot(unittest.TestCase):
    param_file  = "test_tpmsnapshot.params"
    output_file = "test_tidaltensor_tpm_1.0000.hdf5"
    datasources = ['tpm_1.0000.bin.00']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_hdf_result(self, "TidalTensor", rtol=1e-3)



from nbodykit.extensionpoints import DataSource

import numpy
import logging

logger = logging.getLogger('Zheng07Hod')

def set_random_seed(f, seed):
    """
    Decorator to seed the random seed in the 
    ``mc_occuputation_*`` functions of the HOD model instance
    """
    def wrapper(*args, **kwargs):
        kwargs['seed'] = seed
        return f(*args, **kwargs)
    
    for a in f.func_dict:
        setattr(wrapper, a, f.func_dict[a])
    return wrapper
    
    
def model_with_random_seed(model, seed):
    """
    Update the relevant functions of the model to use the
    specified random seed
    """
    # update the satellite profile functions
    sat_functions = ['mc_pos', 'mc_radial_velocity', 'mc_solid_sphere']
    sat_prof = model.model_dictionary['satellites_profile']
    for sat_func in sat_functions:
        f = set_random_seed(getattr(sat_prof, sat_func), seed)
        setattr(sat_prof, sat_func, f)
        
    # update the occupation functions
    occup_functions = ['mc_occupation_centrals', 'mc_occupation_satellites']
    for occup_func in occup_functions:
        f = set_random_seed(getattr(model, occup_func), seed)
        setattr(model, occup_func, f)


class Zheng07HodDataSource(DataSource):
    """
    A `DataSource` that uses the Hod prescription of 
    Zheng et al. 2007 to populate an input halo catalog with galaxies, 
    and returns the (Position, Velocity) of those galaxies
    
    The mock population is done using `halotools` (http://halotools.readthedocs.org)
    The Hod model is of the commonly-used form:
    
    * logMmin: Minimum mass required for a halo to host a central galaxy
    * sigma_logM: Rate of transition from <Ncen>=0 --> <Ncen>=1
    * alpha: Power law slope of the relation between halo mass and <Nsat>
    * logM0: Low-mass cutoff in <Nsat>
    * logM1: Characteristic halo mass where <Nsat> begins to assume a power law form
    
    See the documentation for the `halotools` builtin Zheng07 Hod model, 
    for further details regarding the Hod 
    """
    plugin_name = "Zheng07Hod"
    
    def __init__(self, halocat, redshift, logMmin=13.031, sigma_logM=0.38, 
                    alpha=0.76, logM0=13.27, logM1=14.08, rsd=None, seed=None):
        """
        Default values for Hod values from Reid et al. 2014
        """
        from nbodykit.distributedarray import GatherArray
        
        # load halotools
        try:
            from halotools import sim_manager
        except:
            raise ValueError("`halotools` must be installed to use '%s' DataSource" %self.plugin_name)
            
        # need cosmology
        if self.cosmo is None:
            raise AttributeError("a cosmology instance is required to populate a Hod")
        
        # set global redshift and cosmology as global defaults
        sim_manager.sim_defaults.default_cosmology = self.cosmo.engine
        sim_manager.sim_defaults.default_redshift = self.redshift
        from halotools import empirical_models as em_models
        
        # grab the halocat BoxSize
        self.BoxSize = self.halocat.BoxSize
        
        # read data from halo catalog and then gather to root
        columns = ['Position','Velocity', 'Mass']
        try:
            with self.halocat.open() as stream:
                [data] = stream.read(columns, full=True)
        except Exception as e:
            m = "error reading from halo catalog in %s; " %self.plugin_name
            m += "be sure that the following columns are supported: %s\n" %str(columns)
            raise ValueError(m + "original exception: %s" %str(e))
        alldata = [GatherArray(d, self.comm, root=0) for d in data]
        
        # rank 0 does the populating
        if self.comm.rank == 0:
            Position, Velocity, Mass = alldata
            
            # explicitly set an analytic mass-concentration relation
            sats_prof_model = em_models.NFWPhaseSpace(conc_mass_model='dutton_maccio14')
            
            # build the full hod model, and set our params
            base_model = em_models.PrebuiltHodModelFactory('zheng07')
            self.model = em_models.HodModelFactory(baseline_model_instance=base_model, 
                                                   satellites_profile=sats_prof_model)
        
            for param in self.model.param_dict:
                self.model.param_dict[param] = getattr(self, param)
                
            # set the random seed
            if self.seed is not None:
                model_with_random_seed(self.model, self.seed)
        
            # compute the rsd factor from cosmology
            H = self.cosmo.engine.H(self.redshift) / self.cosmo.engine.h
            a = 1./(1+self.redshift)
            self.rsd_factor = 1./(a*H)  # (velocity in km/s) * rsd_factor = redshift space offset in Mpc/h
        
            # FOFGroups velocity is normalized by aH, so un-normalize into km/s
            ### FIXME
            if self.halocat.plugin_name == 'FOFGroups':
                Velocity /= self.rsd_factor
        
            # make the halotools catalog 
            cols                  = {}
            cols['redshift']      = self.redshift
            cols['Lbox']          = self.halocat.BoxSize.max()
            cols['particle_mass'] = getattr(self.halocat, 'm0', 1.0)
            cols['halo_x']        = Position[:,0]
            cols['halo_y']        = Position[:,1]
            cols['halo_z']        = Position[:,2]
            cols['halo_vx']       = Velocity[:,0]
            cols['halo_vy']       = Velocity[:,1]
            cols['halo_vz']       = Velocity[:,2]
            cols['halo_mvir']     = Mass
            cols['halo_rvir']     = sats_prof_model.halo_mass_to_halo_radius(Mass)
            cols['halo_id']       = numpy.arange(len(Position))
            cols['halo_upid']     = numpy.zeros_like(Mass) - 1
            cols['simname']       = self.plugin_name
            self._halotools_cat   = sim_manager.UserSuppliedHaloCatalog(**cols)
            
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "dataSource for galaxies populating a halo catalog using the Zheng et al. 2007 Hod"
        
        s.add_argument("halocat", type=DataSource.from_config,
            help="DataSource representing the `halo` catalog")
        s.add_argument('redshift', type=float,
            help='the redshift of the ')
        s.add_argument("logMmin", type=float, 
            help="minimum mass required for a halo to host a central galaxy")
        s.add_argument("sigma_logM", type=float, 
            help="rate of transition from <Ncen>=0 --> <Ncen>=1")
        s.add_argument("alpha", type=float, 
            help="power law slope of the relation between halo mass and <Nsat>")
        s.add_argument("logM0", type=float, 
            help="Low-mass cutoff in <Nsat>")
        s.add_argument("logM1", type=float, 
            help="characteristic halo mass where <Nsat> begins to assume a power law form")
        s.add_argument("rsd", type=str, choices="xyz", 
            help="the direction to do the redshift distortion")
        s.add_argument("seed", type=int,
            help='the number used to seed the random number generator')
        
    def log_populated_stats(self):
        """
        Log statistics of the populated catalog
        """
        data = self.model.mock.galaxy_table
        if not len(data):
            raise ValueError("cannot log statistics of an empty galaxy catalog")
            
        logger.info("populated %d galaxies into halo catalog" %len(data))
        fsat = 1.*(data['gal_type'] == 'satellites').sum()/len(data)
        logger.info("  satellite fraction: %.2f" %fsat)
        
        logmass = numpy.log10(data['halo_mvir'])
        logger.info("  mean log10 halo mass: %.2f" %logmass.mean())
        logger.info("  std log10 halo mass: %.2f" %logmass.std())
    
    def update_and_populate(self, **params):
        """
        Update the Hod parameters and populate a mock
        catalog using the new parameters
        
        This is only done on root 0, which is the only rank
        to have the Hod model
        """
        if self.comm.rank == 0:
            
            # update Hod model
            for name in params:
                if name not in self.model.param_dict:
                    valid = list(self.model.param_dict.keys())
                    raise ValueError("'%s' is not a valid Hod parameter name; valid are: %s" %(name, str(valid)))
                self.model.param_dict[name] = params[name]
            
            if not len(self.model.mock.galaxy_table):
                raise ValueError("populated 0 galaxies into the specified halo catalog -- cannot proceed")
                
            # re-populate
            self.populate_mock()
        
    def populate_mock(self):
        """
        Populate the halo catalog with a new set of galaxies. 
        Each call to this function creates a new galaxy catalog,
        overwriting any existing catalog
        
        This is only done on root 0, which is the only rank
        to have the Hod model
        """  
        if self.comm.rank == 0:
            
            # populate for first time (removing only mass=0)
            if not hasattr(self.model, 'mock'):
                self.model.populate_mock(halocat=self._halotools_cat, Num_ptcl_requirement=1)
            # repopulate
            else:
                self.model.mock.populate()
                
            if not len(self.model.mock.galaxy_table):
                raise ValueError("populated 0 galaxies into the specified halo catalog -- cannot proceed")
            self.log_populated_stats()
    
    def parallel_read(self, columns, full=False):
        """
        Return the positions of galaxies, populated into halos using an Hod
        """
        from nbodykit.distributedarray import ScatterArray
        
        # rank 0 does the work, then scatters
        if self.comm.rank == 0:
            
            # populate the model, if need be
            if not hasattr(self.model, 'mock'):
                self.populate_mock()
        
            # the data
            data = self.model.mock.galaxy_table   
        
            # check valid columns
            valid = ['Position', 'Velocity']
            if any(c not in valid and c not in data.colnames for c in columns):
                args = (self.plugin_name, str(valid + data.colnames))
                raise DataSource.MissingColumn('valid column names for %s are: %s' %args)
        
            # position/velocity
            pos = numpy.vstack([data[c] for c in ['x', 'y', 'z']]).T
            vel = numpy.vstack([data[c] for c in ['vx', 'vy', 'vz']]).T
        
            # transform to redshift space
            if self.rsd is not None:
                dir = "xyz".index(self.rsd)
                pos[:,dir] += self.rsd_factor * vel[:,dir] 
        
            toret = []
            for c in columns:
                if c == 'Position':
                    toret.append(pos)
                elif c == 'Velocity':
                    toret.append(vel)
                else:
                    toret.append(numpy.array(data[c].copy()))
        else:
            toret = [None for c in columns]
        
        # scatter the data from rank 0
        newdata = []
        for d in toret:
            newdata.append(ScatterArray(d, self.comm, root=0))

        yield newdata
        
        

            

        
    



from nbodykit.pluginmanager import load_builtins
import os

load_builtins()

# save the path of a few packages
pkg_dir      = os.path.abspath(os.path.join(__file__, '..', '..'))
examples_dir = os.path.join(pkg_dir, 'examples')
bin_dir      = os.path.join(pkg_dir, 'bin')



from nbodykit.extensionpoints import Algorithm
import logging
import numpy
from itertools import product

# for output
import h5py
import bigfile
import mpsort

from mpi4py import MPI

import nbodykit
from pmesh.particlemesh import ParticleMesh
from nbodykit.extensionpoints import DataSource, painters


class TidalTensor(Algorithm):
    plugin_name = "TidalTensor"
    
    def __init__(self, field, points, Nmesh, smoothing=None):
        pass

    @classmethod
    def register(cls):
        s = cls.schema
        s.add_argument("field", type=DataSource.from_config,
                help="DataSource; run `nbkit.py --list-datasources` for all options")
        s.add_argument("points", type=DataSource.from_config,
                help="A small set of points to calculate tidal force on; "
                     "run `nbkit.py --list-datasources` for all options")
        s.add_argument("Nmesh", type=int,
                help='Size of FFT mesh for painting')
        s.add_argument("smoothing", type=float,
                help='Smoothing Length in distance units. '
                      'It has to be greater than the mesh resolution. '
                      'Otherwise the code will die. Default is the mesh resolution.')
                      
    def Smoothing(self, pm, complex):
        k = pm.k
        k2 = 0
        for ki in k:
            ki2 = ki ** 2
            complex *= numpy.exp(-0.5 * ki2 * self.smoothing ** 2)

    def NormalizeDC(self, pm, complex):
        """ removes the DC amplitude. This effectively
            divides by the mean
        """
        w = pm.w
        comm = pm.comm
        ind = []
        value = 0.0
        found = True
        for wi in w:
            if (wi != 0).all():
                found = False
                break
            ind.append((wi == 0).nonzero()[0][0])
        if found:
            ind = tuple(ind)
            value = numpy.abs(complex[ind])
        value = comm.allreduce(value, MPI.SUM)
        complex[:] /= value

    def TidalTensor(self, u, v):
        # k_u k_v / k **2
        def TidalTensor(pm, complex):
            k = pm.k

            for row in range(complex.shape[0]):
                k2 = k[0][row] ** 2
                for ki in k[1:]:
                    k2 = k2 + ki[0] ** 2
                k2[k2 == 0] = numpy.inf
                complex[row] /= k2

            complex *= k[u]
            complex *= k[v]

        return TidalTensor

    def run(self):
        pm = ParticleMesh(self.field.BoxSize, self.Nmesh, dtype='f4', comm=self.comm)
        if self.smoothing is None:
            self.smoothing = self.field.BoxSize[0] / self.Nmesh
        elif (self.field.BoxSize / self.Nmesh > self.smoothing).any():
            raise ValueError("smoothing is too small")
     
        painter = painters.DefaultPainter(weight="Mass")
        painter.paint(pm, self.field)
        pm.r2c()

        pm.transfer([self.Smoothing, self.NormalizeDC])

        with self.points.open() as stream:
            [[Position ]] = stream.read(['Position'], full=True)

        layout = pm.decompose(Position)
        pos1 = layout.exchange(Position)
        value = numpy.empty((3, 3, len(Position)))

        for u, v in product(range(3), range(3)):
            if self.comm.rank == 0:
                logging.info("Working on tensor element (%d, %d)" % (u, v))
            pm.push()
            pm.transfer([self.TidalTensor(u, v)])
            pm.c2r()
            v1 = pm.readout(pos1)
            v1 = layout.gather(v1)
            pm.pop()

            value[u, v] = v1
             
        return value.transpose((2, 0, 1))

    def save(self, output, data):
        self.write_hdf5(data, output)

    def write_hdf5(self, data, output):

        size = self.comm.allreduce(len(data))
        offset = sum(self.comm.allgather(len(data))[:self.comm.rank])

        if self.comm.rank == 0:
            with h5py.File(output, 'w') as ff:
                dataset = ff.create_dataset(name='TidalTensor',
                        dtype=data.dtype, shape=(size, 3, 3))
                dataset.attrs['Smoothing'] = self.smoothing
                dataset.attrs['Nmesh'] = self.Nmesh
                dataset.attrs['Original'] = self.field.string
                dataset.attrs['BoxSize'] = self.field.BoxSize

        for i in range(self.comm.size):
            self.comm.barrier()
            if i != self.comm.rank: continue
                 
            with h5py.File(output, 'r+') as ff:
                dataset = ff['TidalTensor']
                dataset[offset:len(data) + offset] = data




from nbodykit.extensionpoints import Algorithm, DataSource
from nbodykit.distributedarray import GatherArray
import numpy
import logging

class TestBoxSizeAlgorithm(Algorithm):
    """
    A utility algorithm to load a `DataSource` and test 
    if all particles are within the specified BoxSize 
    """
    plugin_name = "TestBoxSize"
    logger = logging.getLogger(plugin_name)

    def __init__(self, datasource, BoxSize):
        pass
        
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "test if all objects in a DataSource fit within a specified BoxSize"

        s.add_argument("datasource", type=DataSource.from_config,
            help="DataSource holding the positions to test")
        s.add_argument("BoxSize", type=DataSource.BoxSizeParser,
            help="the size of the box")
        
                                
    def run(self):
        """
        Run the algorithm, which reads the position coordinates
        from the DataSource and finds any out-of-bounds particles
        """
        # local min/max coords
        local_min_pos = numpy.array([numpy.inf]*3)
        local_max_pos = numpy.array([-numpy.inf]*3)
        
        # read Position
        with self.datasource.open() as stream:
            
            # find global min/max of coords
            for [pos] in stream.read(['Position'], full=False):
                if len(pos):
                    local_min_pos = numpy.minimum(local_min_pos, pos.min(axis=0))
                    local_max_pos = numpy.maximum(local_max_pos, pos.max(axis=0))
                    
            # determine the global min/max
            min_pos = numpy.amin(self.comm.allgather(local_min_pos), axis=0)
            max_pos = numpy.amax(self.comm.allgather(local_max_pos), axis=0)
        
            # and the mean coordinate offset
            self.mean_coordinate_offset = 0.5 * (min_pos + max_pos)
            
            toret = []
            
            # now search for out-of-bounds particles
            for [coords] in stream.read(['Position'], full=False):
                
                pos = coords - self.mean_coordinate_offset
                lim = (pos < -0.5*self.BoxSize)|(pos > 0.5*self.BoxSize)
                idx = lim.any(axis=1)
                
                # the index of each out of bounds particle
                index = idx.nonzero()[0]
                
                # the out-of-bounds boolean 3-vector
                out_of_bounds = lim[idx]
                
                # concat and gather to root
                data = numpy.concatenate([index[:,None], out_of_bounds], axis=1)
                data = GatherArray(data, self.comm, root=0)
                if self.comm.rank == 0:
                    toret.append(data)
                    
            if self.comm.rank == 0:
                toret = numpy.concatenate(toret)
                self.logger.info("%d particles found to be out of range" %len(toret))
                return toret
                
    def save(self, output, result):
        """
        Write out the out-of-bounds particles (done by root)
        """
        # only the master rank writes
        if self.comm.rank == 0:
            
            with open(output, 'wb') as ff:
                
                header = "# %d particles out of bounds\n" %len(result)
                header += "# BoxSize: %s\n" %str(self.BoxSize)
                header += "# mean position vector: %s\n" %str(self.mean_coordinate_offset)
                header += "# column 1: index of out-of-bounds particles (starting at 0)\n"
                header += "# column 2: out of range flag for 'x' dimension (1 if out-of-bounds)\n"
                header += "# column 3: out of range flag for 'y' dimension (1 if out-of-bounds)\n"
                header += "# column 4: out of range flag for 'z' dimension (1 if out-of-bounds)\n"
                ff.write(header.encode())
                numpy.savetxt(ff, result, fmt='%d')

